#
# This source file is part of the EdgeDB open source project.
#
# Copyright 2008-present MagicStack Inc. and the EdgeDB authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#


import typing

from edb import errors

from edb.common import debug
from edb.common import exceptions as edgedb_error

from edb.ir import ast as irast

from edb.pgsql import ast as pgast
from edb.pgsql import codegen as pgcodegen

from . import config as _config_compiler  # NOQA
from . import expr as _expr_compiler  # NOQA
from . import stmt as _stmt_compiler  # NOQA

from . import context
from . import dispatch

from .context import OutputFormat  # NOQA


def compile_ir_to_sql_tree(
        ir_expr: irast.Base, *,
        output_format: typing.Optional[OutputFormat]=None,
        ignore_shapes: bool=False,
        explicit_top_cast: typing.Optional[irast.TypeRef]=None,
        singleton_mode: bool=False,
        use_named_params: bool=False,
        expected_cardinality_one: bool=False) -> pgast.Base:
    try:
        # Transform to sql tree
        ctx_stack = context.CompilerContext()
        ctx = ctx_stack.current
        if isinstance(ir_expr, irast.Statement):
            ctx.scope_tree = ir_expr.scope_tree
            ir_expr = ir_expr.expr
        elif isinstance(ir_expr, irast.ConfigCommand):
            ctx.scope_tree = ir_expr.scope_tree
        ctx.singleton_mode = singleton_mode
        ctx.env = context.Environment(
            output_format=output_format,
            expected_cardinality_one=expected_cardinality_one,
            use_named_params=use_named_params,
            ignore_object_shapes=ignore_shapes,
            explicit_top_cast=explicit_top_cast)
        qtree = dispatch.compile(ir_expr, ctx=ctx)

    except Exception as e:  # pragma: no cover
        try:
            args = [e.args[0]]
        except (AttributeError, IndexError):
            args = []
        raise errors.InternalServerError(*args) from e

    return qtree


def compile_ir_to_sql(
        ir_expr: irast.Base, *,
        output_format: typing.Optional[OutputFormat]=None,
        ignore_shapes: bool=False,
        explicit_top_cast: typing.Optional[irast.TypeRef]=None,
        timer=None,
        use_named_params: bool=False,
        expected_cardinality_one: bool=False,
        pretty: bool=True) -> typing.Tuple[str, typing.Dict[str, int]]:

    if timer is None:
        qtree = compile_ir_to_sql_tree(
            ir_expr,
            output_format=output_format,
            ignore_shapes=ignore_shapes,
            explicit_top_cast=explicit_top_cast,
            use_named_params=use_named_params,
            expected_cardinality_one=expected_cardinality_one)
    else:
        with timer.timeit('compile_ir_to_sql'):
            qtree = compile_ir_to_sql_tree(
                ir_expr,
                output_format=output_format,
                ignore_shapes=ignore_shapes,
                explicit_top_cast=explicit_top_cast,
                use_named_params=use_named_params,
                expected_cardinality_one=expected_cardinality_one)

    if debug.flags.edgeql_compile:  # pragma: no cover
        debug.header('SQL Tree')
        debug.dump(qtree)

    argmap = qtree.argnames

    # Generate query text
    if timer is None:
        codegen = _run_codegen(qtree, pretty=pretty)
    else:
        with timer.timeit('compile_ir_to_sql'):
            codegen = _run_codegen(qtree, pretty=pretty)

    sql_text = ''.join(codegen.result)

    if debug.flags.edgeql_compile:  # pragma: no cover
        debug.header('SQL')
        debug.dump_code(sql_text, lexer='sql')

    return sql_text, argmap


def _run_codegen(qtree, *, pretty=True):
    codegen = pgcodegen.SQLSourceGenerator(pretty=pretty)
    try:
        codegen.visit(qtree)
    except pgcodegen.SQLSourceGeneratorError as e:  # pragma: no cover
        ctx = pgcodegen.SQLSourceGeneratorContext(
            qtree, codegen.result)
        edgedb_error.add_context(e, ctx)
        raise
    except Exception as e:  # pragma: no cover
        ctx = pgcodegen.SQLSourceGeneratorContext(
            qtree, codegen.result)
        err = pgcodegen.SQLSourceGeneratorError(
            'error while generating SQL source')
        edgedb_error.add_context(err, ctx)
        raise err from e

    return codegen
